# -*- coding: utf-8 -*-

import os
import errno
import signal
import logging

import psutil
from noseapp.utils.common import waiting_for
from noseapp.utils.common import TimeoutException


logger = logging.getLogger(__name__)


def safe_shot_down(process, timeout=3.0, sleep=0.5):
    """
    :type process: psutil.Popen
    """
    try:

        try:
            children = (c for c in process.children(recursive=True))
            map(safe_shot_down, children)
        except AttributeError:
            pass

        process.terminate()

        try:
            waiting_for(
                process.poll,
                sleep=sleep,
                timeout=timeout,
            )
        except TimeoutException:
            process.kill()

    except (psutil.NoSuchProcess, OSError):
        pass


def process_terminate_by_pid_file(pid_file):
    """
    :type pid_file: PidFileObject
    """
    if not pid_file.exist:
        return

    try:
        os.kill(pid_file.pid, signal.SIGTERM)
    except OSError:
        pass


class PidFileObject(object):

    def __init__(self, file_path):
        """
        :param file_path: pid file path
        """
        self._file_path = file_path

    @property
    def exist(self):
        try:
            return os.path.isfile(self._file_path)
        except TypeError:
            return False

    @property
    def path(self):
        return self._file_path

    @property
    def pid(self):
        if not self._file_path:
            return None

        try:
            with open(self._file_path) as fp:
                return int(fp.readline().strip())
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise
        except ValueError:
            pass

        return None

    def remove(self):
        if self.exist:
            try:
                os.unlink(self._file_path)
            except OSError:
                pass

    def __repr__(self):
        return '<PidFile {}>'.format(self._file_path)



# -*- coding: utf-8 -*-

from collections import OrderedDict
from contextlib import contextmanager

from noseapp_daemon.runner import DaemonRunner
from noseapp_daemon.service import DaemonService


class ServiceNotFound(BaseException):
    pass


class DaemonNotFound(BaseException):
    pass


class DaemonManagement(object):
    """
    Interface for daemons and services control

    Example::

      class MyDaemonManagement(DaemonManagement):

        def setup(self):
          # do something...

      management = MyDaemonManagement(app)
      management.add_daemon(daemon)
      ...

      with management.checkout_daemon(
        'my_daemon',
        ignore_exc=(ValueError,),
        callback=lambda: None,  # if ValueError will be raised, this function can be called
      ) as my_daemon:
        my_daemon.start()

      management.service('my_service').start()

      management.stop_all()

      management.start_all()
      management.stop_daemons()
      ...
    """

    def __init__(self, app=None, **options):
        self._app = app
        self._options = options

        self.__daemons = OrderedDict()
        self.__services = OrderedDict()

        self.setup()

    def setup(self):
        pass

    def add_service(self, service):
        if not isinstance(service, DaemonService):
            raise TypeError('"service" param is not instance "DaemonService"')

        self.__services[service.name] = service

    def add_daemon(self, daemon):
        if not isinstance(daemon, DaemonRunner):
            raise TypeError('"daemon" param is not instance "DaemonRunner"')

        self.__daemons[daemon.name] = daemon

    def daemon(self, name):
        try:
            daemon = self.__daemons[name]
        except KeyError:
            raise DaemonNotFound(name)

        return daemon

    def service(self, name):
        try:
            service = self.__services[name]
        except KeyError:
            raise ServiceNotFound(name)

        return service

    @contextmanager
    def checkout_service(self, name, ignore_exc=None, callback=None):
        if isinstance(ignore_exc, tuple):
            try:
                yield self.service(name)
            except ignore_exc:
                if callable(callback):
                    callback()
        else:
            yield self.service(name)

    @contextmanager
    def checkout_daemon(self, name, ignore_exc=None, callback=None):
        if isinstance(ignore_exc, tuple):
            try:
                yield self.daemon(name)
            except ignore_exc:
                if callable(callback):
                    callback()
        else:
            yield self.daemon(name)

    def start_services(self):
        for _, service in self.__services.items():
            service.start()

    def stop_services(self):
        for _, service in self.__services.items():
            service.stop()

    def restart_services(self):
        self.stop_services()
        self.start_services()

    def start_daemons(self):
        for _, daemon in self.__daemons.items():
            daemon.start()

    def stop_daemons(self):
        for _, daemon in self.__daemons.items():
            daemon.stop()

    def restart_daemons(self):
        self.stop_daemons()
        self.start_daemons()

    def start_all(self):
        self.start_daemons()
        self.start_services()

    def stop_all(self):
        self.stop_daemons()
        self.stop_services()

    def restart_all(self):
        self.stop_all()
        self.start_all()



# -*- coding: utf-8 -*-

import logging
from subprocess import PIPE

import psutil

from noseapp_daemon import utils


logger = logging.getLogger(__name__)


class CallbackInterface(object):
    """
    Callback methods for daemon management
    """

    @property
    def process_options(self):
        """
        Arguments for Popen __init__ method
        """
        return {}

    def setup(self):
        """
        Be called after initialization
        """
        pass

    def before_start(self):
        """
        Pre start callback
        """
        pass

    def after_start(self):
        """
        Post start callback
        """
        pass

    def before_stop(self):
        """
        Pre stop callback
        """
        pass

    def after_stop(self):
        """
        Post stop callback
        """
        pass


class DaemonRunner(CallbackInterface):
    """
    Base class for daemon run
    """

    def __init__(self,
                 daemon_bin=None,
                 stdout=None,
                 stderr=None,
                 pid_file=None,
                 cmd_prefix=None,
                 **options):
        """
        :param daemon_bin: path to executable file
        :type daemon_bin: str
        :param pid_file: path to pid file
        :type pid_file: str
        :param cmd_prefix: pre start command line prefix
        :type cmd_prefix: basestring, tuple
        :param stdout: path to stdout log file.
        :type stdout: str
        :param stderr: path to stderr log file.
        :type stderr: str
        """
        self.daemon_bin = daemon_bin
        self.pid_file = utils.PidFileObject(pid_file)
        self.options = options

        if isinstance(cmd_prefix, basestring):
            self.cmd_prefix = [c.strip() for c in cmd_prefix.split(' ')]
        else:
            self.cmd_prefix = cmd_prefix

        self.process = None

        self.stdout = stdout
        self.stderr = stderr

        self.setup()

    @property
    def name(self):
        raise NotImplementedError(
            'Property "name" should be implemented in subclasses.',
        )

    @property
    def cmd(self):
        raise NotImplementedError(
            'Property "cmd" should be implemented in subclasses.',
        )

    @property
    def started(self):
        return bool(self.process) or self.pid_file.exist

    @property
    def stopped(self):
        return not self.started

    def start(self):
        """
        Start daemon
        """
        if not self.process and self.pid_file.exist:
            self.pid_file.remove()

        if self.started:
            return

        logger.debug('Daemod "%s" start', self.name)

        self.before_start()

        self.process = psutil.Popen(
            self.cmd if not self.cmd_prefix else self.cmd_prefix + self.cmd,
            stderr=PIPE,
            stdout=PIPE,
            **self.process_options
        )

        self.after_start()

    def stop(self):
        """
        Stop daemon
        """
        if self.stopped:
            return

        logger.debug('Daemod "%s" stop', self.name)

        self.before_stop()

        utils.process_terminate_by_pid_file(self.pid_file)
        utils.safe_shot_down(self.process)

        self.process.wait()

        self.pid_file.remove()

        if self.stdout:
            with open(self.stdout, 'a') as stdout:
                for line in self.process.stdout:
                    stdout.write(line)

        if self.stderr:
            with open(self.stderr, 'a') as stderr:
                for line in self.process.stderr:
                    stderr.write(line)

        self.process = None

        self.after_stop()



# -*- coding: utf-8 -*-

from noseapp_daemon.service import DaemonService
from noseapp_daemon.management import DaemonManagement
from noseapp_daemon.runner import DaemonRunner as Daemon


__version__ = '1.0.0'


__all__ = (
    Daemon,
    DaemonService,
    DaemonManagement,
)



# -*- coding: utf-8 -*-


class DaemonService(object):
    """
    Abstract layer for daemon run logic realization

    Example::

      class MyDaemonService(DaemonService):

        def setup(self):
          self._daemon = MyDaemon(self._config...)
          # do something...

        # your interface for daemon management here

        def start():
          self._daemon.start()
    """

    def __init__(self, config=None, **options):
        self._config = config
        self._options = options

        self.setup()

    def setup(self):
        pass

    @property
    def name(self):
        raise NotImplementedError(
            'Property "name" should be implemented in subclasses.',
        )

    def start(self):
        raise NotImplementedError(
            'Method "start" should be implemented in subclasses.',
        )

    def stop(self):
        raise NotImplementedError(
            'Method "stop" should be implemented in subclasses.',
        )

    def restart(self):
        self.stop()
        self.start()



# -*- coding: utf-8 -*-

from setuptools import setup
from setuptools import find_packages

import noseapp_daemon


if __name__ == '__main__':
    setup(
        name='noseapp_daemon',
        url='https://github.com/trifonovmixail/noseapp_daemon',
        version=noseapp_daemon.__version__,
        packages=find_packages(),
        author='Mikhail Trifonov',
        author_email='mikhail.trifonov@corp.mail.ru',
        description='noseapp extension for linux daemon testing',
        include_package_data=True,
        zip_safe=False,
        platforms='linux',
        install_requires=[
            'noseapp',
            'psutil==2.2.1',
        ],
    )
